/* -*- mode: C; tab-width: 2; indent-tabs-mode: nil; -*- */
/* Table Toy Benchmark (11/26/97 version):
 *
 * The benchmark is successfully completed if at least 99% of the table
 * is correctly built; it *is* OK, for example, to have a few errors
 * enter into the table due to rare race conditions between processors
 * updating the same memory location.
 *
 * It is expected that in a multi-processor version, one will start the
 * random number generator at N equally spaced spots along its cycle.
 * Each processor then steps the generator through its section of the
 * cycle.  (The starts routine provided can be used to start the
 * random number generator at any specified point.)
 *
 * The benchmark objective is to measure interprocessor bandwidth.  It
 * is ok to bucket sort locally and use a message passing protocol
 * between processors if this is faster than a global shared memory
 * approach.  More sophisticated optimizations which attempt to reorder
 * the loop so that all updates are local are not considered "in the
 * spirit" of the benchmark.
 *
 */

#include <hpcc.h>

#include "RandomAccess.h"

#define PROCS (1L << LPROCS)

/* #define LEXTBK 1 best*/
#define LEXTBK 1
/* #define LEXTBK 0 */
#define EXTBK (1L << LEXTBK) /* buckets per processor */

/* Log size of main table (suggested: half of global memory) */

#ifndef LTABSIZE
/* 32 MB */
#define LTABSIZE 25
#endif

#ifdef LONG_IS_64BITS
#define TABSIZE (1L << LTABSIZE)
#else
#define TABSIZE (1LL << LTABSIZE)
#endif

/* Number of updates to table (suggested: 4x number of table entries) */
#define NUPDATE (4 * TableSize)

/* Allocate main table (in global memory) */
/*unsigned long *Table;*/
u64Int *Table;

#define TRUE 1
#define FALSE 0
#define DONE 0
#define SLOT_CNT 1
#define FIRST_SLOT 2

/* Local buckets to sort into */
#define BK_SZ 2048
/* #define BK_SZ 32768 */
/* #define BK_SZ 16384 */
/* #define BK_SZ 8192 */
/* `LocalBuckets' is used as 2D array - two last dimensions are considered
 together. In the call to MPI_Alltoall() both `LocalBuckets' and
 `GlobalBuckets' are assumed to be stored in a contiguous area of memory. */
/*
u64Int LocalBuckets[PROCS][EXTBK][(BK_SZ+2)];
u64Int GlobalBuckets[PROCS][EXTBK*(BK_SZ+2)];
*/
u64Int *LocalBuckets;
u64Int *GlobalBuckets;
/* 2D and 3D array accessors */
#define A2D0(a, i, j, d2) a[(size_t)(i) * (d2) + (j)]
#define A2D(a, i, j) a[(size_t)(i) * EXTBK * (BK_SZ+2) + (j)]
#define A3D0(a, i, j, k, d2, d3) a[((size_t)(i) * (d2) + (j)) * (d3) + (k)]
#define A3D(a, i, j, k) a[((size_t)(i) * EXTBK + (j)) * (BK_SZ+2) + (k)]

/* Log size of substitution table (suggested: half of primary cache) */
#define LSTSIZE 9
#define STSIZE (1 << LSTSIZE)

u64Int Stable[STSIZE];        /* Substitution table */
u64Int Ran;                   /* Current random numbers */
/* int PEUpdateDone[PROCS]; */
/* s64Int PEUpdateDone[PROCS]; */
s64Int *PEUpdateDone;

void
Sum64(void *invec, void *inoutvec, int *len, MPI_Datatype *datatype) {
  int i, n = *len; s64Int *invec64 = invec, *inoutvec64 = inoutvec;
  for (i = n; i; i--, invec64++, inoutvec64++) *inoutvec64 += *invec64;
}

int
MPIRandomAccess(HPCC_Params *params) {
  s64Int i;
  s64Int send_cnt,glb_send_cnt, k, l;
/*  int numprocs,myproc;
  int next_slot, which_pe, m,sub_buk;
  int lc_recv_done =  FALSE;
  int gap, bsz, tail, mindata;
*/
  int numprocs,myproc;
  s64Int next_slot, which_pe, m,sub_buk;
  s64Int lc_recv_done =  FALSE;
  s64Int gap, bsz, tail, mindata;

  u64Int logTableSize, TableSize;
  u64Int temp, ran_tmp;
  double cputime;               /* CPU  time to update table */
  double realtime;              /* Real time to update table */
  double totalMem;
  int sAbort, rAbort;
  long logNumProcs, NumProcs;
  int ii, ij, color, key;
  FILE *outFile;
  MPI_Comm comm; MPI_Op sum64;
  double *GUPs;

  GUPs = &params->MPIGUPs;

  MPI_Comm_size( MPI_COMM_WORLD, &numprocs );
  MPI_Comm_rank( MPI_COMM_WORLD, &myproc );

  if (0 == myproc) {
    outFile = fopen( params->outFname, "a" );
    if (! outFile) outFile = stderr;
  }

  sAbort = 0; if (sizeof(u64Int) < 8) sAbort = 1;

  MPI_Allreduce( &sAbort, &rAbort, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD );
  if (rAbort > 0) {
    if (myproc == 0) fprintf( outFile, "No 64-bit integer type available.\n" );
    goto comp_end;
  }

  totalMem = params->HPLMaxProcMem; /* max single node memory */
  totalMem *= params->HPLMaxProc;   /* max all-node memory */
  totalMem /= sizeof(u64Int);

  /* calculate the size of update array (must be a power of 2) */
  for (totalMem *= 0.5, logTableSize = 0, TableSize = 1;
       totalMem >= 1.0;
       totalMem *= 0.5, logTableSize++, TableSize <<= 1)
    ; /* EMPTY */

  /* calculate the number of processors (must be a power of 2) */
  for (ii = 1, logNumProcs = 0; ; logNumProcs++, ii <<= 1) {
    if (ii == numprocs) { /* number of processes is a power 2 */
      NumProcs = ii;
      comm = MPI_COMM_WORLD;
      break;

    /* number of processes is not a power 2 (too many shifts may introduce negative values or 0) */
    } else if (ii > numprocs || ii <= 0) {
      NumProcs = ij;
      logNumProcs--;
      if (myproc < NumProcs) color = 0;
      else color = MPI_UNDEFINED;
      key = myproc;
      MPI_Comm_split( MPI_COMM_WORLD, color, key, &comm );
      break;

    } else { /* (ii < numprocs) -- `ii' is still too small */
      ij = ii;
    }
  }

  sAbort = 0;
  /* if at least one process doesn't receive a communicator but it should */
  if (MPI_COMM_NULL == comm && myproc < NumProcs) sAbort = 1;

  MPI_Allreduce( &sAbort, &rAbort, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD );
  if (rAbort > 0) {
    if (myproc == 0) fprintf( outFile, "Could not create a new communicator.\n" );
    if (MPI_COMM_NULL != comm) goto failed_table;
  }

  if (MPI_COMM_NULL == comm) goto comp_end;

  Table = XMALLOC( u64Int, TableSize/NumProcs );
  sAbort = 0; if (! Table) sAbort = 1;

  MPI_Allreduce( &sAbort, &rAbort, 1, MPI_INT, MPI_SUM, comm );
  if (rAbort > 0) {
    if (myproc == 0) fprintf( outFile, "Failed to allocate memory for the main table.\n" );
    goto failed_table;
  }

  /* u64Int LocalBuckets[PROCS][EXTBK][(BK_SZ+2)]; */
  LocalBuckets = XMALLOC( u64Int, NumProcs * EXTBK * (BK_SZ+2) );
  sAbort = 0; if (! LocalBuckets) sAbort = 1;

  MPI_Allreduce( &sAbort, &rAbort, 1, MPI_INT, MPI_SUM, comm );
  if (rAbort > 0) {
    if (myproc == 0) fprintf( outFile, "Failed to allocate memory for LocalBuckets.\n" );
    goto failed_local;
  }

  /* u64Int GlobalBuckets[PROCS][EXTBK*(BK_SZ+2)]; */
  GlobalBuckets = XMALLOC( u64Int, NumProcs * EXTBK * (BK_SZ+2) );
  sAbort = 0; if (! GlobalBuckets) sAbort = 1;

  MPI_Allreduce( &sAbort, &rAbort, 1, MPI_INT, MPI_SUM, comm );
  if (rAbort > 0) {
    if (myproc == 0) fprintf( outFile, "Failed to allocate memory for GlobalBuckets.\n" );
    goto failed_global;
  }

  /* s64Int PEUpdateDone[PROCS]; */
  PEUpdateDone = XMALLOC( s64Int, NumProcs );
  sAbort = 0; if ( ! PEUpdateDone) sAbort = 1;

  MPI_Allreduce( &sAbort, &rAbort, 1, MPI_INT, MPI_SUM, comm );
  if (rAbort > 0) {
    if (myproc == 0) fprintf( outFile, "Failed to allocate memory for PEUpdateDone.\n" );
    goto failed_done;
  }

  if (myproc == 0) {
    fprintf( outFile, "Running on %ld processors\n", NumProcs);
    fprintf( outFile, "Total Main table size = 2^" FSTR64 " = " FSTR64 " words\n",
             logTableSize, TableSize );
    fprintf( outFile, "PE Main table size = 2^" FSTR64 " = " FSTR64 " words/PE\n",
             logTableSize - logNumProcs, TableSize/NumProcs );
    fprintf( outFile, "Subst table size = 2^%d = %d words\n", LSTSIZE, STSIZE);
    fprintf( outFile, "Number of updates = " FSTR64 "\n", NUPDATE);
  }

  /* Initialize substitution table (not time critical) */
  Stable[0] = 0;
  for (i=1; i<STSIZE; i++)
    Stable[i] = Stable[i-1] +
#ifdef LONG_IS_64BITS
      0x0123456789abcdefL;
#else
      0x0123456789abcdefLL;
#endif

  MPI_Barrier( comm );

  /* Begin timing here */
  cputime = -CPUSEC();
  realtime = -RTSEC();

  /* Initialize main table */
  for (i=0; i<TableSize/NumProcs; i++)
    Table[i] = i;

  /* Perform updates to main table.  The scalar equivalent is:
   *
   *     u64Int Ran;
   *     Ran = 1;
   *     for (i=0; i<NUPDATE; i++) {
   *       Ran = (Ran << 1) ^ (((s64Int) Ran < 0) ? POLY : 0);
   *       Table[Ran & (TABSIZE-1)] ^= Stable[Ran >> (64-LSTSIZE)];
   *     }
   */

  Ran = starts ((NUPDATE/NumProcs) * myproc);

  for (i=0; i<NumProcs; i++)
    PEUpdateDone[i] = FALSE;

  send_cnt = NUPDATE/NumProcs;

  while(lc_recv_done == FALSE){
    if (send_cnt > 0) {
      /* Initalize local buckets */
      for (i=0; i<NumProcs; i++){
	for (l=0; l<EXTBK; l++){
	  /*LocalBuckets[i][l][SLOT_CNT] = FIRST_SLOT;*/
	  A3D( LocalBuckets, i, l, SLOT_CNT ) = FIRST_SLOT;
	  /* LocalBuckets[i][l][DONE] = FALSE; */
	  A3D( LocalBuckets, i, l, DONE ) = FALSE;
	}
      }
      /* Fill local buckets until one is full or out of data */
      next_slot = FIRST_SLOT;
      while(next_slot<BK_SZ+2 && send_cnt>0 ) {
	Ran =(Ran << 1) ^ ((s64Int) Ran < ZERO64B ? POLY : ZERO64B);
	which_pe = (Ran >> (logTableSize - logNumProcs)) & (NumProcs - 1);
        sub_buk  = (Ran >> (logTableSize - (logNumProcs + LEXTBK)) & (EXTBK - 1));
	next_slot = A3D( LocalBuckets, which_pe, sub_buk, SLOT_CNT );
        A3D( LocalBuckets, which_pe, sub_buk, next_slot ) = Ran;
	A3D( LocalBuckets, which_pe, sub_buk, SLOT_CNT ) = ++next_slot;
	send_cnt--;
      }

      if (send_cnt == 0)
	for (i=0; i<NumProcs; i++)
	  for (l=0; l<EXTBK; l++)
	    A3D( LocalBuckets, i, l, DONE ) = TRUE;
#if LEXTBK > 0
      /* Compress the buckets */
      for(k=0; k<NumProcs; k++){
	/* For each destination compress the EXTBK buckets into one
	   bucket for sending.*/
	for (l=1;l<EXTBK;l++){
	  /* Find the size of the gap between buckets */
	  gap = (l*(BK_SZ+2)+2) - A3D( LocalBuckets, k, 0, SLOT_CNT );
          /* Find the ammout of data in the bucket to be moved */
	  bsz = A3D( LocalBuckets, k, l, SLOT_CNT ) - 2;
          /* Calculate the amount of data that needs to be moved to
	     have a single bucket */
	  mindata = gap < bsz ? gap : bsz;
	  /* Calculate where the data is coming from */
	  tail = l*(BK_SZ+2) + A3D( LocalBuckets, k, l, SLOT_CNT ) - 1;
	  /* Calculate where the data is going to */
	  next_slot = A3D( LocalBuckets, k, 0, SLOT_CNT );
	  /* #pragma  unroll  32 */
	  /* Move the data */
	  for (m=tail; m>(tail-mindata); m--){
	    A3D( LocalBuckets, k, 0, next_slot ) = A3D( LocalBuckets, k, 0, m );
	    next_slot++;
	  }

	  A3D( LocalBuckets, k, 0, SLOT_CNT ) =  next_slot;

	  if (gap < bsz)
	    A3D( LocalBuckets, k, 0, SLOT_CNT ) += bsz - gap;
	}
      } /* End of compress loop */
#endif
    } /* End of sending loop */

    MPI_Barrier( comm );

    lc_recv_done = TRUE;

    /* Now move all the buckets to the appropriate pe*/

#ifdef LONG_IS_64BITS
    MPI_Alltoall(LocalBuckets,  EXTBK*(BK_SZ+2), MPI_UNSIGNED_LONG,
		 GlobalBuckets, EXTBK*(BK_SZ+2), MPI_UNSIGNED_LONG,
		 comm);
#else
    MPI_Alltoall(LocalBuckets,  EXTBK*(BK_SZ+2), MPI_LONG_LONG_INT,
		 GlobalBuckets, EXTBK*(BK_SZ+2), MPI_LONG_LONG_INT,
		 comm);
#endif

    for(k=0; k<NumProcs; k++){
      /* Each Pe updates its part of the table */
      /* if we didn't see the end of the data yet update table with latest */
      if(PEUpdateDone[k] == FALSE){
	PEUpdateDone[k] = A2D( GlobalBuckets, k, DONE );
	/* #pragma unroll  3 */
	for(m=2; m < A2D( GlobalBuckets, k, SLOT_CNT ); m++){
          ran_tmp = A2D( GlobalBuckets, k, m );
	  Table[(ran_tmp & ((1<<(logTableSize - logNumProcs))-1))]
	    ^= Stable[ran_tmp >> (64-LSTSIZE)];
	}
	lc_recv_done &= PEUpdateDone[k];
      }
    }

  } /* no more local data */

  /* End timed section */
  cputime += CPUSEC();
  realtime += RTSEC();

  /* Print timing results */
  if (myproc == 0){
    *GUPs = 10e-9*NUPDATE / realtime;
    fprintf( outFile, "CPU time used = %.6f seconds\n", cputime );
    fprintf( outFile, "Real time used = %.6f seconds\n", realtime );
    fprintf( outFile, "%.9f Billion(10^9) Updates    per second [GUP/s]\n", *GUPs );
    fprintf( outFile, "%.9f Billion(10^9) Updates/PE per second [GUP/s]\n",
             *GUPs / NumProcs );
  }
  MPI_Bcast( GUPs, 1, MPI_INT, 0, comm ); /* distribute result to all nodes */

  /* Verification of results (in serial or "safe" mode; optional) */
#if 1
  cputime  = -CPUSEC();
  realtime = -RTSEC();
  temp = 0x1;
  for (i=0; i<NUPDATE; i++) {
    temp = (temp << 1) ^ (((s64Int) temp < ZERO64B) ? POLY : ZERO64B);
    if(((temp >> (logTableSize - logNumProcs)) & (NumProcs - 1)) == myproc)
      Table[temp & ((1 <<  (logTableSize - logNumProcs)) -1 )]
	^= Stable[temp >> (64-LSTSIZE)];
  }

  send_cnt = 0;
  for (i=0; i<TableSize/NumProcs; i++)
    if (Table[i] != i)
      send_cnt++;

#ifdef LONG_IS_64BITS
  MPI_Reduce( &send_cnt, &glb_send_cnt, 1, MPI_LONG, MPI_SUM, 0, comm );
#else
  /* MPI 1.1 standard (obsolete at this point) doesn't define MPI_SUM to work on `long long':
    http://www.mpi-forum.org/docs/mpi-11-html/node78.html
    and therefore LAM 6.5.6 chooses not to implement it (even though there is code for it in LAM and
    for other reductions work OK, e.g. MPI_MAX). MPICH 1.2.5 doesn't complain about MPI_SUM but it
    doesn't have MPI_UNSIGNED_LONG_LONG (but has MPI_LONG_LONG_INT):
    http://www.mpi-forum.org/docs/mpi-20-html/node84.htm
    So I need to create a trivial summation operation. */
  MPI_Op_create( Sum64, 1, &sum64 );
  MPI_Reduce( &send_cnt, &glb_send_cnt, 1, MPI_LONG_LONG_INT, sum64, 0, comm );
  MPI_Op_free( &sum64 );
#endif

  cputime += CPUSEC();
  realtime += RTSEC();
  if(myproc == 0){
    fprintf( outFile, "Check  CPU time used = %.6f seconds\n", cputime);
    fprintf( outFile, "Check Real time used = %.6f seconds\n", realtime);
    fprintf( outFile, "Found " FSTR64 " errors in " FSTR64 " locations (%s).\n",
	    glb_send_cnt, TableSize, (glb_send_cnt <= 0.01*TableSize) ?
	    "passed" : "failed");
  }
#endif

  /* Deallocate memory (in reverse order of allocation which should help fragmentation) */
  free( PEUpdateDone );

  failed_done:

  free( GlobalBuckets );

  failed_global:

  free( LocalBuckets );

  failed_local:

  free( Table );

  failed_table:

  if (MPI_COMM_WORLD != comm) MPI_Comm_free( &comm );

  comp_end:

  if (0 == myproc) if (outFile != stderr) fclose( outFile );

  MPI_Barrier( MPI_COMM_WORLD );

  return 0;
}
